import pygame as pg

class Toolbar(object):

    def __init__(self, screen, items = ['Form', 'Textbox']):
        self.screen = screen
        self.items = items
        self.item_width = 180
        self.item_height = 30
        self.font = pg.font.SysFont('Arial', 26)

    def draw(self):
        pg.draw.rect(self.screen, (0, 0, 0), (0, 0, self.item_width, self.item_height), 2)
        main_text = self.font.render('Items:', False, (0, 0, 0))
        w, h = main_text.get_rect().width, main_text.get_rect().height
        self.screen.blit(main_text, (self.item_width//2 - w//2, 0))

        for i, item in enumerate(self.items):
            pg.draw.rect(self.screen, (0, 0, 0), (0, (i + 1)*self.item_height, self.item_width, self.item_height), 2)
            item_text = self.font.render(item, False, (0, 0, 0))
            w, h = item_text.get_rect().width, item_text.get_rect().height
            self.screen.blit(item_text, (self.item_width // 2 - w // 2, (i + 1)*self.item_height))

    def find_button(self, x, y):
        if x > self.item_width or x < 0:
            return None
        for i, item in enumerate(self.items):
            if (i + 1)*self.item_height <= y < (i + 2)*self.item_height:
                return item
        return None


import pygame as pg

class ExceptionHandler(object):

    def __init__(self, screen):
        self.screen = screen
        self.message = ''
        self.font = pg.font.SysFont('Arial', 20)

    def draw(self):
        to_display = self.message.split('\n')
        cur_h = 0
        init_height_offset = 200
        width, height = self.screen.get_rect().width, self.screen.get_rect().height
        for i, line in enumerate(to_display):
            text = self.font.render(line, False, (0, 0, 0))
            w, h = text.get_rect().width, text.get_rect().height
            self.screen.blit(text, (5, height - init_height_offset + cur_h))
            cur_h += h
        pg.draw.line(self.screen, (0, 0, 0), (0, height - init_height_offset - 5), (width, height - init_height_offset - 5), 2)



import pygame as pg
from .utils import init_events, check_name_eligibility, str_to_tuple, find_font
from .base_icon import BaseIcon

class Label(BaseIcon):

    defaults = {'type' : 'Label',
                'name' : None,
                'width' : 100,
                'height' : 30,
                'x' : None,
                'y' : None,
                'text' : 'Label',
                'font_type' : 'arial',
                'font_size' : 20,
                'font_color' : [0, 0, 0],
                'enabled' : True,
                'visible' : True}

    updated = defaults.copy()

    def __init__(self, form, x, y, exception_handler):
        super().__init__(exception_handler)
        self.__dict__.update(self.defaults)
        self.form = form
        self.x = x
        self.y = y
        self.abs_x = x + self.form.x
        self.abs_y = y + self.form.y
        self.events = init_events()

    def draw(self):
        self.font = pg.font.SysFont(self.font_type, self.font_size)
        main_text = self.font.render(self.text, False, self.font_color)
        if self.visible:
            self.form.surface.blit(main_text, (self.x, self.y))
        w, h = main_text.get_rect().width, main_text.get_rect().height
        if self.is_selected:
            pg.draw.rect(self.form.surface, (255, 0, 0), (self.x, self.y, w, h), 1)
        self.abs_x = self.x + self.form.x
        self.abs_y = self.y + self.form.y
        self.width = w
        self.height = h

    def copy(self):
        copied = Label(self.form, self.x, self.y, self.exception_handler)
        for key in self.updated:
            copied.__dict__[key] = self.__dict__[key]
        return copied


import pygame as pg
from .utils import init_events, check_name_eligibility, str_to_tuple, find_font
from .base_icon import BaseIcon

class Form(BaseIcon):

    defaults = {'type' : 'Form',
                 'name' : None,
                 'width' : 700,
                 'height' : 700,
                 'x' : None,
                 'y' : None,
                 'bg_color' : [211, 211, 211]}

    updated = defaults.copy()

    def __init__(self, screen, x, y, exception_handler):
        super().__init__(exception_handler)
        self.__dict__.update(self.defaults)
        self.screen = screen
        self.x = x
        self.y = y
        self.updated['x'], self.updated['y'] = x, y
        self.objects = []
        self.events = init_events()


    def draw(self):
        self.surface = pg.Surface((self.width, self.height))
        if self.is_selected:
            self.surface.fill((255, 0, 0))
        else:
            self.surface.fill((0, 0, 0))
        pg.draw.rect(self.surface, self.bg_color, (2, 2, self.width-4, self.height-4))
        self.draw_objects()
        self.screen.blit(self.surface, (self.x, self.y))

    def draw_objects(self):
        for object in self.objects:
            object.draw()

    def copy(self):
        copied = Form(self.screen, self.x, self.y, self.exception_handler)
        for key in self.updated:
            copied.__dict__[key] = self.__dict__[key]
        if self.is_selected:
            copied.is_selected = True
        for obj in self.objects:
            new_obj = obj.copy()
            if obj.is_selected:
                new_obj.is_selected = True
            new_obj.form = copied
            copied.objects.append(new_obj)
        return copied



import pygame as pg

def init_events():
    events = ['mouse_button_left_down', 'mouse_button_left_up', 'backspace_down', 'enter_down', 'space_down', 'delete_down',
              'mouse_button_right_down', 'mouse_button_right_up', 'cCTRL_down', 'vCTRL_down', 'sCTRL_down']
    for i in range(33, 127):
        events.append(chr(i) + '_down')
        events.append(chr(i).upper() + '_down')
    events_dict = {}
    for event in events:
        events_dict[event] = None
    return events_dict

def check_name_eligibility(name, form):
    if form.name == name:
        raise Exception
    for object in form.objects:
        if object.name == name:
            raise Exception
    return name

def str_to_tuple(data):
    data = str(data)
    conv_data = ''
    for char in data:
        if char != ' ':
            conv_data += char
    if (conv_data[0] != '[' and conv_data[0] != '(') or (conv_data[-1] != ']' and conv_data[-1] != ')'):
        raise Exception
    conv_data = conv_data[1:-1]
    conv_data = conv_data.split(',')
    if len(conv_data) != 3:
        raise Exception
    for i, value in enumerate(conv_data):
        conv_data[i] = int(value)
        if int(value) < 0 or int(value) > 255:
            raise Exception
    return conv_data

def find_font(font):
    if font in pg.font.get_fonts():
        return font
    else:
        raise Exception






import pygame as pg
from .utils import init_events, check_name_eligibility, str_to_tuple, find_font
from .base_icon import BaseIcon

class Canvas(BaseIcon):

    defaults = {'type' : 'Canvas',
                'name' : None,
                'width' : 200,
                'height' : 200,
                'x' : None,
                'y' : None,
                'bg_color' : [255, 255, 255],
                'enabled' : True,
                'visible' : True}

    updated = defaults.copy()

    def __init__(self, form, x, y, exception_handler):
        super().__init__(exception_handler)
        self.__dict__.update(self.defaults)
        self.form = form
        self.x = x
        self.y = y
        self.abs_x = x + self.form.x
        self.abs_y = y + self.form.y
        self.events = init_events()
        self.surface = None
        self.draw_function = None

    def draw(self):
        self.surface = pg.Surface((self.width, self.height))
        if self.is_selected:
            self.surface.fill((255, 0, 0))
        else:
            self.surface.fill((0, 0, 0))
        pg.draw.rect(self.surface, self.bg_color, (1, 1, self.width - 2, self.height - 2))
        if self.draw_function is not None:
            self.draw_function()
        self.form.surface.blit(self.surface, (self.x, self.y))
        self.abs_x = self.x + self.form.x
        self.abs_y = self.y + self.form.y

    def copy(self):
        copied = Canvas(self.form, self.x, self.y, self.exception_handler, )
        for key in self.updated:
            copied.__dict__[key] = self.__dict__[key]
        return copied




import pygame as pg
from .utils import init_events, check_name_eligibility, str_to_tuple, find_font
from .base_icon import BaseIcon

class Button(BaseIcon):

    defaults = {'type' : 'Button',
                'name' : None,
                'width' : 100,
                'height' : 30,
                'x' : None,
                'y' : None,
                'text' : 'Button',
                'max_len' : 10,
                'bg_color' : [169, 169, 169],
                'font_type' : 'arial',
                'font_size' : 24,
                'font_color' : [0, 0, 0],
                'enabled' : True,
                'visible' : True}

    updated = defaults.copy()

    def __init__(self, form, x, y, exception_handler):
        super().__init__(exception_handler)
        self.__dict__.update(self.defaults)
        self.form = form
        self.x = x
        self.y = y
        self.abs_x = x + self.form.x
        self.abs_y = y + self.form.y
        self.events = init_events()


    def draw(self):
        self.font = pg.font.SysFont(self.font_type, self.font_size)
        self.conv_text = self.text.split('\n')
        surface = pg.Surface((self.width, self.height))
        if self.is_selected:
            surface.fill((255, 0, 0))
        else:
            surface.fill((0, 0, 0))
        pg.draw.rect(surface, self.bg_color, (1, 1, self.width - 2, self.height - 2))
        tot_h = 0
        for line in self.conv_text:
            main_text = self.font.render(line, False, self.font_color)
            w, h = main_text.get_rect().width, main_text.get_rect().height
            surface.blit(main_text, (self.width // 2 - w//2, tot_h))
            tot_h += h
        if self.visible:
            self.form.surface.blit(surface, (self.x, self.y))
        self.abs_x = self.x + self.form.x
        self.abs_y = self.y + self.form.y

    def copy(self):
        copied = Button(self.form, self.x, self.y, self.exception_handler)
        for key in self.updated:
            copied.__dict__[key] = self.__dict__[key]
        return copied


import pygame as pg
from .utils import init_events, check_name_eligibility, str_to_tuple, find_font

class BaseIcon(object):

    defaults = {'type': None,
                'name': None,
                'width': None,
                'height': None,
                'x': None,
                'y': None}

    updated = defaults.copy()

    def __init__(self, exception_handler):
        self.__dict__.update(self.defaults)
        self.events = init_events()
        self.exception_handler = exception_handler
        self.is_selected = False
        self.times_copied = 0

    def add_event(self, event, callback):
        if event not in self.events:
            raise Exception(f'{event} is not a correct event')
        else:
            self.events[event] = callback

    def call_event(self, event):
        if self.events[event] is not None:
            return self.events[event]()

    def update_dictionary(self):
        for key in self.updated:
            self.updated[key] = self.__dict__[key]

    def convert_variables(self):
        error = False
        if self.updated['width'] == '':
            self.updated['width'] = '0'
        if self.updated['height'] == '':
            self.updated['height'] = '0'
        if self.updated['x'] == '':
            self.updated['x'] = '0'
        if self.updated['y'] == '':
            self.updated['y'] = '0'

        if self.updated['type'] != self.type:
            error = True
            self.exception_handler.message = 'Changing the type of an icon is impossible.'

        try:
            if self.type != 'Form':
                self.name = check_name_eligibility(self.updated['name'], self.form)
                self.name = self.updated['name']
            else:
                self.name = check_name_eligibility(self.updated['name'], self)
                self.name = self.updated['name']
        except:
            if self.name != self.updated['name']:
                error = True
                self.exception_handler.message = 'The following name: {} already exists in the project.'.format(self.updated['name'])
        try:
            self.width = int(self.updated['width'])
            self.width = max(0, self.width)
        except:
            error = True
            self.exception_handler.message = 'Width must be a positive integer.'
        try:
            self.height = int(self.updated['height'])
            self.height = max(0, self.height)
        except:
            error = True
            self.exception_handler.message = 'Height must be a positive integer.'
        try:
            self.x = int(self.updated['x'])
            self.x = max(0, self.x)
        except:
            error = True
            self.exception_handler.message = 'x coordinate must be a positive integer.'
        try:
            self.y = int(self.updated['y'])
            self.y = max(0, self.y)
        except:
            error = True
            self.exception_handler.message = 'y coordinate must be a positive integer.'
        if 'max_len' in self.updated:
            try:
                self.max_len = int(self.updated['max_len'])
            except:
                error = True
                self.exception_handler.message = 'The maximum text length must be a positive integer.'
        if 'text' in self.updated:
            try:
                if len(self.updated['text']) <= self.max_len:
                    self.text = self.updated['text']
                else:
                    self.text = self.updated['text'][:self.max_len]
            except:
                error = True
                self.exception_handler.message = 'Failed to update text.'
        if 'text_align' in self.updated:
            try:
                if self.updated['text_align'] in ['left-top', 'middle-top', 'right-top',
                                                  'left-middle', 'middle', 'right-middle',
                                                  'left-down', 'middle-down', 'right-down']:
                    self.text_align = self.updated['text_align']
                else:
                    raise Exception
            except:
                error = True
                self.exception_handler.message = 'This text align is not supported. The following is a list of all text-aligns:\n' \
                                                 'left-top, middle-top, right-top\n' \
                                                 'left-middle, middle, right-middle\n' \
                                                 'left-down, middle-down, right-down'
        if 'bg_color' in self.updated:
            try:
                self.bg_color = str_to_tuple(self.updated['bg_color'])
            except:
                error = True
                self.exception_handler.message = 'This is not a correct color value. Color should be input with the following scheme: [R, G, B] where:\n' \
                                                 'R is an integer from 0 to 255 signifying color red,\n' \
                                                 'G is an integer from 0 to 255 signifying color green,\n' \
                                                 'B is an integer from 0 to 255 signifying color blue.'
        if 'font_type' in self.updated:
            try:
                self.font_type = find_font(self.updated['font_type'])
            except:
                error = True
                self.exception_handler.message = '{} is not an existing pygame font. Run pygame.font.get_fonts() to see available fonts.'.format(self.updated['font_type'])
        if 'font_size' in self.updated:
            try:
                self.font_size = int(self.updated['font_size'])
                self.font_size = max(0, self.font_size)
            except:
                error = True
                self.exception_handler.message = 'Font size must be a positive integer.'
        if 'font_color' in self.updated:
            try:
                self.font_color = str_to_tuple(self.updated['font_color'])
            except:
                error = True
                self.exception_handler.message = 'This is not a correct color value. Color should be input with the following scheme: [R, G, B] where:\n' \
                                                 'R is an integer from 0 to 255 signifying color red,\n' \
                                                 'G is an integer from 0 to 255 signifying color green,\n' \
                                                 'B is an integer from 0 to 255 signifying color blue.'
        if 'enabled' in self.updated:
            try:
                if self.updated['enabled'] in [True, 'true', 'True', '1']:
                    self.enabled = True
                elif self.updated['enabled'] in [False, 'false', 'False', '0']:
                    self.enabled = False
                else:
                    raise Exception
            except:
                error = True
                self.exception_handler.message = 'This is an incorrect boolean type. In order to set a boolean variable, type:\n' \
                                                 'For True: True, true or 1,\n' \
                                                 'For False: False, false or 0.'
        if 'visible' in self.updated:
            try:
                if self.updated['visible'] in [True, 'true', 'True', '1']:
                    self.visible = True
                elif self.updated['visible'] in [False, 'false', 'False', '0']:
                    self.visible = False
                else:
                    raise Exception
            except:
                error = True
                self.exception_handler.message = 'This is an incorrect boolean type. In order to set a boolean variable, type:\n' \
                                                 'For True: True, true or 1,\n' \
                                                 'For False: False, false or 0.'
        return error





import pygame as pg

class Memory(object):

    def __init__(self, max_len = 20):
        self.max_len = max_len
        self.memory = []
        self.memory_index = -1

    def remember(self, form):
        if self.memory_index >= 0:
            self.memory = self.memory[:self.memory_index + 1]
        self.memory.append(form)
        if len(self.memory) > self.max_len:
            self.memory.pop(0)
        self.memory_index += 1

    def move_record(self, action):
        if len(self.memory) > 0:
            self.memory_index += action
            self.memory_index = min(len(self.memory) - 1, self.memory_index)
            self.memory_index = max(0, self.memory_index)
            return self.memory[self.memory_index]


import pygame as pg

class Listener(object):

    def __init__(self):
        pass

    def listen(self):
        mods = pg.key.get_mods()
        for event in pg.event.get():
            if event.type == pg.QUIT:
                return 'quit'
            elif event.type == pg.KEYDOWN:
                if event.key == pg.K_ESCAPE:
                    return 'quit'
                elif event.key == pg.K_s and mods & pg.KMOD_CTRL:
                    return 'sCTRL_down'
                elif event.key == pg.K_c and mods & pg.KMOD_CTRL:
                    return 'cCTRL_down'
                elif event.key == pg.K_v and mods & pg.KMOD_CTRL:
                    return 'vCTRL_down'
                elif event.key == pg.K_z and mods & pg.KMOD_CTRL:
                    return 'zCTRL_down'
                elif event.key == pg.K_y and mods & pg.KMOD_CTRL:
                    return 'yCTRL_down'
                elif 33 <= event.key <= 126 and not (mods & pg.KMOD_SHIFT or mods & pg.KMOD_CAPS):
                    return chr(event.key) + '_down'
                elif 33 <= event.key <= 126 and (mods & pg.KMOD_SHIFT or mods & pg.KMOD_CAPS):
                    return chr(event.key).upper() + '_down'
                elif event.key == pg.K_BACKSPACE:
                    return 'backspace_down'
                elif event.key == pg.K_SPACE:
                    return 'space_down'
                elif event.key == pg.K_RETURN:
                    return 'enter_down'
                elif event.key == pg.K_DELETE:
                    return 'delete_down'
            elif event.type == pg.MOUSEBUTTONDOWN:
                if event.button == 1:
                    return 'mouse_button_left_down'
                elif event.button == 3:
                    return 'mouse_button_right_down'
            elif event.type == pg.MOUSEBUTTONUP:
                if event.button == 1:
                    return 'mouse_button_left_up'
                elif event.button == 3:
                    return 'mouse_button_right_up'
        return None


# Loading the libraries
import pygame as pg
import os

# Loading the icons
from .objects.form import Form
from .objects.textbox import Textbox
from .objects.button import Button
from .objects.label import Label
from .objects.canvas import Canvas

# Loading tools
from .tools.specification_viewer import SpecificationViewer
from .tools.listener import Listener
from .tools.toolbar import Toolbar
from .tools.memory import Memory
from .tools.exception_handler import ExceptionHandler


class FormLoader(object):

    def __init__(self, save_filepath, screen=None):
        pg.init()
        self.save_filepath = save_filepath
        self.screen = screen
        self.listener = Listener()
        self.exception_handler = ExceptionHandler(self.screen)
        self.form = None
        self.objects = {}
        self.item_selected = None

    def draw(self):
        self.screen.fill((0, 0, 0))
        if self.form is not None:
            self.form.draw()
        pg.display.flip()

    def detect_object(self, x, y):
        for obj in self.form.objects:
            if obj.abs_x <= x <= obj.abs_x + obj.width and obj.abs_y <= y <= obj.abs_y + obj.height:
                return obj
        if self.form.x <= x <= self.form.x + self.form.width and self.form.y <= y <= self.form.y + self.form.height:
            return self.form
        return None

    def update(self):
        self.draw()
        event = self.listener.listen()
        x, y = pg.mouse.get_pos()
        if event == 'quit':
            pg.quit()
        elif event == 'mouse_button_left_down':
            object_detected = self.detect_object(x, y)
            if object_detected is None:
                self.item_selected.is_selected = False
                self.item_selected = None
            else:
                if self.item_selected is not None:
                    self.item_selected.is_selected = False
                self.item_selected = object_detected
        elif event == 'backspace_down':
            if self.item_selected is not None and self.item_selected.type == 'Textbox' and len(
                    self.item_selected.text) > 0:
                if self.item_selected.enabled:
                    self.item_selected.text = self.item_selected.text[:-1]
        elif event == 'enter_down':
            if self.item_selected is not None and self.item_selected.type == 'Textbox' and len(
                    self.item_selected.text) < self.item_selected.max_len:
                if self.item_selected.enabled:
                    self.item_selected.text += '\n'
        elif event == 'space_down':
            if self.item_selected is not None and self.item_selected.type == 'Textbox' and len(
                    self.item_selected.text) < self.item_selected.max_len:
                if self.item_selected.enabled:
                    self.item_selected.text += ' '
        elif event is not None and len(event) == 6 and event[-4:] == 'down':
            if self.item_selected is not None and self.item_selected.type == 'Textbox' and len(
                    self.item_selected.text) < self.item_selected.max_len:
                if self.item_selected.enabled:
                    self.item_selected.text += event[0]

        if self.item_selected is not None:
            self.item_selected.is_selected = True
            if event is not None:
                return self.item_selected.call_event(event)

    def load(self):
        data_separator = '##END DATA##'
        savefile = open(self.save_filepath, 'r').read().splitlines()
        loaded_objects = []
        cur_object = {}
        for line in savefile:
            if line == data_separator:
                loaded_objects.append(cur_object)
                cur_object = {}
            else:
                key, value = line.split(':')
                if key == 'text':
                    value = value.replace('#', '\n')
                cur_object[key] = value
        for object_data in loaded_objects:
            if object_data['type'] == 'Form':
                x, y = int(object_data['x']), int(object_data['y'])
                os.environ['SDL_VIDEO_WINDOW_POS'] = "%d,%d" % (x, y)
                if self.screen is None:
                    self.screen = pg.display.set_mode((int(object_data['width']), int(object_data['height'])))
                    self.exception_handler.screen = self.screen
                    self.form = Form(self.screen, 0, 0, self.exception_handler)
                    object_data['x'] = object_data['y'] = 0
                else:
                    self.form = Form(self.screen, int(object_data['x']), int(object_data['y']), self.exception_handler)
                self.form.updated.update(object_data)
                self.form.convert_variables()
                self.objects[object_data['name']] = self.form
            elif object_data['type'] == 'Button':
                x, y = int(object_data['x']), int(object_data['y'])
                button = Button(self.form, x, y, self.exception_handler)
                button.updated.update(object_data)
                button.convert_variables()
                self.form.objects.append(button)
                self.objects[object_data['name']] = button
            elif object_data['type'] == 'Textbox':
                x, y = int(object_data['x']), int(object_data['y'])
                textbox = Textbox(self.form, x, y, self.exception_handler)
                textbox.updated.update(object_data)
                textbox.convert_variables()
                self.form.objects.append(textbox)
                self.objects[object_data['name']] = textbox
            elif object_data['type'] == 'Label':
                x, y = int(object_data['x']), int(object_data['y'])
                label = Label(self.form, x, y, self.exception_handler)
                label.updated.update(object_data)
                label.convert_variables()
                self.form.objects.append(label)
                self.objects[object_data['name']] = label
            elif object_data['type'] == 'Canvas':
                x, y = int(object_data['x']), int(object_data['y'])
                canvas = Canvas(self.form, x, y, self.exception_handler)
                canvas.updated.update(object_data)
                canvas.convert_variables()
                self.form.objects.append(canvas)
                self.objects[object_data['name']] = canvas
        print('Form Loaded')


import pygame as pg

class SpecificationViewer(object):

    def __init__(self, screen):
        self.screen = screen
        self.cell_width = 180
        self.cell_height = 30
        self.font = pg.font.SysFont('Arial', 26)
        self.displayed_object = None

    def draw(self, item):
        self.displayed_object = item
        w, h = self.screen.get_rect().width, self.screen.get_rect().height
        pg.draw.rect(self.screen, (0, 0, 0), (w - 2 * self.cell_width, 0, self.cell_width, self.cell_height), 2)
        pg.draw.rect(self.screen, (0, 0, 0), (w - self.cell_width, 0, self.cell_width, self.cell_height), 2)
        field_text = self.font.render('Field', False, (0, 0, 0))
        tw1, th1 = field_text.get_rect().width, field_text.get_rect().height
        self.screen.blit(field_text, (w - 3*self.cell_width//2 - tw1//2, self.cell_height//2 - th1//2))
        value_text = self.font.render('Value', False, (0, 0, 0))
        tw2, th2 = value_text.get_rect().width, value_text.get_rect().height
        self.screen.blit(value_text, (w - self.cell_width//2 - tw2//2, self.cell_height//2 - th2//2))

        for i, (key, value) in enumerate(item.updated.items()):
            pg.draw.rect(self.screen, (0, 0, 0), (w - 2 * self.cell_width, (i+1)*self.cell_height, self.cell_width, self.cell_height), 2)
            pg.draw.rect(self.screen, (0, 0, 0), (w - self.cell_width, (i+1)*self.cell_height, self.cell_width, self.cell_height), 2)
            field_text = self.font.render(str(key), False, (0, 0, 0))
            tw1, th1 = field_text.get_rect().width, field_text.get_rect().height
            self.screen.blit(field_text, (w - 3 * self.cell_width // 2 - tw1 // 2, self.cell_height // 2 - th1 // 2 + (i+1)*self.cell_height))
            value_text = self.font.render(str(value), False, (0, 0, 0))
            tw2, th2 = value_text.get_rect().width, value_text.get_rect().height
            if tw2 < self.cell_width - 4:
                self.screen.blit(value_text, (w - self.cell_width // 2 - tw2 // 2, self.cell_height // 2 - th2 // 2 + (i+1)*self.cell_height))
            else:
                value_text = self.font.render('--too long--', False, (0, 0, 0))
                tw2, th2 = value_text.get_rect().width, value_text.get_rect().height
                self.screen.blit(value_text, (
                w - self.cell_width // 2 - tw2 // 2, self.cell_height // 2 - th2 // 2 + (i + 1) * self.cell_height))

    def find_field(self, x, y):
        if self.displayed_object is not None:
            w = self.screen.get_rect().width
            if x < w - 2*self.cell_width:
                return None
            for i, (field, value) in enumerate(self.displayed_object.updated.items()):
                if (i + 1)*self.cell_height <= y <= (i + 2)*self.cell_height:
                    return field
            return None




import pygame as pg
from .utils import init_events, check_name_eligibility, str_to_tuple, find_font
from .base_icon import BaseIcon

class Textbox(BaseIcon):

    defaults = {'type' : 'Textbox',
                 'name' : None,
                 'width' : 100,
                 'height' : 30,
                 'x' : None,
                 'y' : None,
                 'text' : 'Textbox',
                 'text_align' : 'left-top',
                 'max_len' : 10,
                 'bg_color' : [255, 255, 255],
                 'font_type' : 'arial',
                 'font_size' : 24,
                 'font_color' : [0, 0, 0],
                 'enabled' : True,
                 'visible' : True}

    updated = defaults.copy()

    def __init__(self, form, x, y, exception_handler):
        super().__init__(exception_handler)
        self.__dict__.update(self.defaults)
        self.form = form
        self.x = x
        self.y = y
        self.abs_x = x + self.form.x
        self.abs_y = y + self.form.y
        self.conv_text = self.text.split('\n')
        self.events = init_events()

    def draw(self):
        self.font = pg.font.SysFont(self.font_type, self.font_size)
        self.conv_text = self.text.split('\n')
        surface = pg.Surface((self.width, self.height))
        if self.is_selected:
            surface.fill((255, 0, 0))
        else:
            surface.fill((0, 0, 0))
        pg.draw.rect(surface, self.bg_color, (1, 1, self.width - 2, self.height - 2))
        tot_h = 0
        text_to_blit = []
        for line in self.conv_text:
            main_text = self.font.render(line, False, self.font_color)
            w, h = main_text.get_rect().width, main_text.get_rect().height
            text_to_blit.append(main_text)
            tot_h += h

        if self.text_align == 'left-top':
            cur_h = 0
            for line in text_to_blit:
                h = line.get_rect().height
                surface.blit(line, (5, cur_h))
                cur_h += h
        elif self.text_align == 'middle-top':
            cur_h = 0
            for line in text_to_blit:
                w, h = line.get_rect().width, line.get_rect().height
                surface.blit(line, (self.width//2 - w//2, cur_h))
                cur_h += h
        elif self.text_align == 'right-top':
            cur_h = 0
            for line in text_to_blit:
                w, h = line.get_rect().width, line.get_rect().height
                surface.blit(line, (self.width - 5 - w, cur_h))
                cur_h += h
        elif self.text_align == 'left-middle':
            cur_h = 0
            for line in text_to_blit:
                h = line.get_rect().height
                surface.blit(line, (5, self.height//2 - tot_h//2 + cur_h))
                cur_h += h
        elif self.text_align == 'middle':
            cur_h = 0
            for line in text_to_blit:
                w, h = line.get_rect().width, line.get_rect().height
                surface.blit(line, (self.width//2 - w//2, self.height//2 - tot_h//2 + cur_h))
                cur_h += h
        elif self.text_align == 'right-middle':
            cur_h = 0
            for line in text_to_blit:
                w, h = line.get_rect().width, line.get_rect().height
                surface.blit(line, (self.width - w - 5, self.height//2 - tot_h//2 + cur_h))
                cur_h += h
        elif self.text_align == 'left-down':
            cur_h = 0
            for line in text_to_blit:
                h = line.get_rect().height
                surface.blit(line, (5, self.height - tot_h + cur_h))
                cur_h += h
        elif self.text_align == 'middle-down':
            cur_h = 0
            for line in text_to_blit:
                w, h = line.get_rect().width, line.get_rect().height
                surface.blit(line, (self.width//2 - w//2, self.height - tot_h + cur_h))
                cur_h += h
        elif self.text_align == 'right-down':
            cur_h = 0
            for line in text_to_blit:
                w, h = line.get_rect().width, line.get_rect().height
                surface.blit(line, (self.width - w - 5, self.height - tot_h + cur_h))
                cur_h += h

        if self.visible:
            self.form.surface.blit(surface, (self.x, self.y))
        self.abs_x = self.x + self.form.x
        self.abs_y = self.y + self.form.y

    def copy(self):
        copied = Textbox(self.form, self.x, self.y, self.exception_handler)
        for key in self.updated:
            copied.__dict__[key] = self.__dict__[key]
        return copied


# Loading the libraries
import pygame as pg
import os

# Loading the icons
from .objects.form import Form
from .objects.textbox import Textbox
from .objects.button import Button
from .objects.label import Label
from .objects.canvas import Canvas

# Loading tools
from .tools.specification_viewer import SpecificationViewer
from .tools.listener import Listener
from .tools.toolbar import Toolbar
from .tools.memory import Memory
from .tools.exception_handler import ExceptionHandler

from .form_loader import FormLoader


class FormCreator(object):

    def __init__(self, save_filepath, load_from_save=True):
        pg.init()
        self.save_filepath = save_filepath
        self.screen = pg.display.set_mode((0, 0), pg.FULLSCREEN)
        self.listener = Listener()
        self.specification_viewer = SpecificationViewer(self.screen)
        self.exception_handler = ExceptionHandler(self.screen)
        self.memory = Memory()
        self.items = ['Form', 'Textbox', 'Button', 'Label', 'Canvas']
        self.items_count = {}
        for item in self.items:
            self.items_count[item] = 0
        self.toolbar = Toolbar(self.screen, items=self.items)
        self.form = None
        self.item_held = None
        self.prev_pos_left_down = (None, None)
        self.init_pos_right_down = (None, None)
        self.resize_action = None
        self.item_resized = None
        self.group_selected = []
        self.to_copy = None
        self.has_changed = False
        self.last_pattern_pos = (None, None)
        self.value_change = None
        if os.path.exists(self.save_filepath) and load_from_save:
            self.load()
            if self.form is not None:
                for obj in self.form.objects:
                    self.items_count[obj.type] += 1

    def detect_rect_edge(self, x, y, w, h):
        mx, my = pg.mouse.get_pos()
        tolerance = 3
        if x - tolerance <= mx <= x + tolerance and y - tolerance <= my <= y + tolerance:
            return 0
        elif x + tolerance <= mx <= x + w - tolerance and y - tolerance <= my <= y + tolerance:
            return 1
        elif x + w - tolerance <= mx <= x + w + tolerance and y - tolerance <= my <= y + tolerance:
            return 2
        elif x + w - tolerance <= mx <= x + w + tolerance and y + tolerance <= my <= y + h - tolerance:
            return 3
        elif x + w - tolerance <= mx <= x + w + tolerance and y + h - tolerance <= my <= y + h + tolerance:
            return 4
        elif x + tolerance <= mx <= x + w - tolerance and y + h - tolerance <= my <= y + h + tolerance:
            return 5
        elif x - tolerance <= mx <= x + tolerance and y + h - tolerance <= my <= y + h + tolerance:
            return 6
        elif x - tolerance <= mx <= x + tolerance and y + tolerance <= my <= y + h - tolerance:
            return 7
        return None

    def mouse_in_item(self, item, is_on_form=False):
        x, y = pg.mouse.get_pos()
        if not is_on_form and item.x <= x <= item.x + item.width and item.y <= y <= item.y + item.height:
            return True
        elif is_on_form and item.abs_x <= x <= item.abs_x + item.width and item.abs_y <= y <= item.abs_y + item.height:
            return True
        return False

    def left_mouse_down(self):
        return pg.mouse.get_pressed()[0] == 1

    def draw_objects(self):
        self.toolbar.draw()
        self.exception_handler.draw()
        if self.form is not None:
            self.form.draw()

        if len(self.group_selected) == 1:
            self.specification_viewer.draw(self.group_selected[0])
        else:
            self.specification_viewer.displayed_object = None

    def update_screen(self):
        self.screen.fill((255, 255, 255))
        self.draw_objects()
        if self.init_pos_right_down != (None, None):
            x, y = self.init_pos_right_down
            mx, my = pg.mouse.get_pos()
            pg.draw.rect(self.screen, (0, 0, 0), (x, y, mx - x, my - y), 1)
        pg.display.flip()

    def find_selected_group(self):
        if self.init_pos_right_down != (None, None):
            for obj in self.group_selected:
                obj.is_selected = False
            self.group_selected = []
            x, y = self.init_pos_right_down
            mx, my = pg.mouse.get_pos()
            if x > mx:
                x, mx = mx, x
            if y > my:
                y, my = my, y
            w, h = mx - x, my - y
            for obj in self.form.objects:
                if x <= obj.abs_x <= x + w and y <= obj.abs_y <= y + h:
                    self.group_selected.append(obj)

    def create_new_object(self, type):
        x, y = pg.mouse.get_pos()
        item_to_create = None
        if type == 'Form' and self.form is None:
            self.form = Form(self.screen, 300, 100, self.exception_handler)
            item_to_create = self.form
            self.has_changed = False
            self.form.name = type + str(self.items_count[type])
            self.items_count[type] += 1
        elif type == 'Textbox' and self.form is not None:
            textbox = Textbox(self.form, 100, 100, self.exception_handler)
            textbox.name = type + str(self.items_count[type])
            textbox.name = self.get_unique_name(textbox)
            item_to_create = textbox
            self.form.objects.append(textbox)
            self.has_changed = False
            self.items_count[type] += 1
        elif type == 'Button' and self.form is not None:
            button = Button(self.form, 100, 120, self.exception_handler)
            button.name = type + str(self.items_count[type])
            button.name = self.get_unique_name(button)
            item_to_create = button
            self.form.objects.append(button)
            self.has_changed = False
            self.items_count[type] += 1
        elif type == 'Label' and self.form is not None:
            label = Label(self.form, 100, 140, self.exception_handler)
            label.name = type + str(self.items_count[type])
            label.name = self.get_unique_name(label)
            item_to_create = label
            self.form.objects.append(label)
            self.has_changed = False
            self.items_count[type] += 1
        elif type == 'Canvas' and self.form is not None:
            canvas = Canvas(self.form, 100, 160, self.exception_handler)
            canvas.name = type + str(self.items_count[type])
            canvas.name = self.get_unique_name(canvas)
            item_to_create = canvas
            self.form.objects.append(canvas)
            self.has_changed = False
            self.items_count[type] += 1

        if item_to_create is not None:
            self.group_selected = [item_to_create]
            item_to_create.is_selected = True
            self.memory.remember(self.form.copy())
            if item_to_create.type != 'Form':
                return item_to_create

    def get_unique_name(self, item):
        unique = True
        if self.form.name == item.name:
            unique = False
        for obj in self.form.objects:
            if obj.name == item.name:
                unique = False
        if unique:
            return item.name
        else:
            max_inx = 0
            for obj in self.form.objects:
                if obj.type == item.type and len(obj.name) > len(obj.type) and obj.name[:len(obj.type)] == obj.type:
                    inx = obj.name[len(obj.type):]
                    try:
                        inx = int(inx)
                        if inx > max_inx:
                            max_inx = inx
                    except:
                        pass
            self.items_count[obj.type] = max_inx + 1
            return item.type + str(max_inx + 1)

    def copy_group(self):
        if self.form not in self.to_copy:
            for obj in self.group_selected:
                obj.is_selected = False
            self.group_selected = []
            for obj in self.to_copy:
                if obj.type != 'Form':
                    new_obj = obj.copy()
                    new_obj.x = obj.x + 10 * (obj.times_copied + 1)
                    new_obj.y = obj.y + 10 * (obj.times_copied + 1)
                    new_obj.name = new_obj.type + str(self.items_count[new_obj.type])
                    new_obj.name = self.get_unique_name(new_obj)
                    self.form.objects.append(new_obj)
                    self.group_selected.append(new_obj)
                    new_obj.is_selected = True
                    obj.times_copied += 1
                    self.items_count[new_obj.type] += 1
            self.memory.remember(self.form.copy())
            self.has_changed = False

    def resize_item(self, item):
        if self.resize_action is not None and self.left_mouse_down():
            x, y = pg.mouse.get_pos()
            prev_x, prev_y = self.prev_pos_left_down
            diffx, diffy = x - prev_x, y - prev_y
            if self.resize_action == 0 and item.width - diffx >= 0 and item.height - diffy >= 0:
                item.x += diffx
                item.y += diffy
                item.width -= diffx
                item.height -= diffy
                self.has_changed = True
            elif self.resize_action == 1 and item.height - diffy >= 0:
                item.y += diffy
                item.height -= diffy
                self.has_changed = True
            elif self.resize_action == 2 and item.height - diffy >= 0 and item.width + diffx >= 0:
                item.width += diffx
                item.y += diffy
                item.height -= diffy
                self.has_changed = True
            elif self.resize_action == 3 and item.width + diffx >= 0:
                item.width += diffx
                self.has_changed = True
            elif self.resize_action == 4 and item.width + diffx >= 0 and item.height + diffy >= 0:
                item.width += diffx
                item.height += diffy
                self.has_changed = True
            elif self.resize_action == 5 and item.height + diffy >= 0:
                item.height += diffy
                self.has_changed = True
            elif self.resize_action == 6 and item.width - diffx >= 0 and item.height + diffy >= 0:
                item.x += diffx
                item.width -= diffx
                item.height += diffy
                self.has_changed = True
            elif self.resize_action == 7 and item.width - diffx >= 0:
                item.x += diffx
                item.width -= diffx
                self.has_changed = True
            self.prev_pos_left_down = pg.mouse.get_pos()

    def run(self):
        while True:
            event = self.listener.listen()
            if event == 'quit':
                pg.quit()
            elif event == 'mouse_button_left_down':
                self.value_change = None
                self.prev_pos_left_down = pg.mouse.get_pos()
                x, y = self.prev_pos_left_down
                if self.form is not None and not (self.form.x <= x <= self.form.x + self.form.width and self.form.y <= y <= self.form.y + self.form.height):
                    if self.toolbar.item_width <= x <= self.screen.get_rect().width - 2 * self.specification_viewer.cell_width:
                        for obj in self.group_selected:
                            obj.update_dictionary()
                            obj.is_selected = False
                        self.group_selected = []
                    elif self.screen.get_rect().width - 2 * self.specification_viewer.cell_width <= x <= self.screen.get_rect().width and len(self.group_selected) == 1:
                        for obj in self.group_selected:
                            obj.update_dictionary()
                        self.value_change = self.specification_viewer.find_field(x, y)
                    else:
                        for obj in self.group_selected:
                            obj.update_dictionary()
                            obj.is_selected = False
                        self.group_selected = []
                        if self.item_held is not None:
                            self.item_held.is_selected = False
                            self.item_held = None

                if self.item_held is None:
                    item = self.toolbar.find_button(x, y)
                    if item is not None:
                        self.create_new_object(item)
                else:
                    self.item_held = None
            elif event == 'mouse_button_left_up':
                self.resize_action = None
                self.item_held = None
                self.item_resized = None
                if self.has_changed:
                    self.memory.remember(self.form.copy())
                    self.has_changed = False
            elif event == 'mouse_button_right_down':
                x, y = pg.mouse.get_pos()
                self.init_pos_right_down = (x, y)
            elif event == 'mouse_button_right_up':
                self.init_pos_right_down = (None, None)
            elif event == 'backspace_down':
                if self.value_change is None and len(self.group_selected) == 1 is not None and self.group_selected[0].type == 'Textbox' and len(self.group_selected[0].text) > 0:
                    self.group_selected[0].text = self.group_selected[0].text[:-1]
                    self.memory.remember(self.form.copy())
                    self.has_changed = False
                elif self.value_change is not None and len(self.group_selected) == 1:
                    value = self.group_selected[0].updated[self.value_change]
                    if len(str(value)) > 0:
                        self.group_selected[0].updated[self.value_change] = str(value)[:-1]
            elif event == 'space_down':
                if self.value_change is None and len(self.group_selected) == 1 and self.group_selected[0].type == 'Textbox' and len(self.group_selected[0].text) < self.group_selected[0].max_len:
                    self.group_selected[0].text += ' '
                    self.memory.remember(self.form.copy())
                    self.has_changed = False
                elif self.value_change is not None and len(self.group_selected) == 1:
                    self.group_selected[0].updated[self.value_change] = str(
                        self.group_selected[0].updated[self.value_change]) + ' '
            elif event == 'enter_down':
                if self.value_change is None and len(self.group_selected) == 1 and self.group_selected[0].type == 'Textbox' and len(self.group_selected[0].text) < self.group_selected[0].max_len:
                    self.group_selected[0].text += '\n'
                    self.memory.remember(self.form.copy())
                    self.has_changed = False
                elif self.value_change is not None and len(self.group_selected) == 1:
                    error = self.group_selected[0].convert_variables()
                    self.group_selected[0].update_dictionary()
                    if not error:
                        self.memory.remember(self.form.copy())
                    self.has_changed = False
            elif event == 'delete_down' and len(self.group_selected) != 0:
                for obj in self.group_selected:
                    if obj is not self.form:
                        self.form.objects.remove(obj)
                self.group_selected = []
                self.memory.remember(self.form.copy())
                self.has_changed = False
            elif event == 'sCTRL_down':
                self.save()
                self.exception_handler.message = 'Form saved.'
            elif event == 'cCTRL_down':
                self.to_copy = self.group_selected
            elif event == 'vCTRL_down':
                self.copy_group()
            elif event == 'zCTRL_down':
                self.form = self.memory.move_record(-1).copy()
                self.group_selected = []
                if self.form.is_selected:
                    self.group_selected = [self.form]
                for obj in self.form.objects:
                    if obj.is_selected:
                        self.group_selected.append(obj)
                self.item_held = None
                self.item_resized = None
                self.value_change = None
                self.resize_action = None
                self.to_copy = []
            elif event == 'yCTRL_down':
                self.form = self.memory.move_record(1).copy()
                self.group_selected = []
                if self.form.is_selected:
                    self.group_selected = [self.form]
                for obj in self.form.objects:
                    if obj.is_selected:
                        self.group_selected.append(obj)
                self.item_held = None
                self.item_resized = None
                self.value_change = None
                self.resize_action = None
                self.to_copy = []
            elif event is not None and len(event) == 6 and event[-4:] == 'down':
                if self.value_change is None and len(self.group_selected) == 1 and self.group_selected[0].type == 'Textbox' and len(self.group_selected[0].text) < self.group_selected[0].max_len:
                    self.group_selected[0].text += event[0]
                    self.memory.remember(self.form.copy())
                    self.has_changed = False
                elif self.value_change is not None and len(self.group_selected) == 1:
                    self.group_selected[0].updated[self.value_change] = str(
                        self.group_selected[0].updated[self.value_change]) + event[0]

            if self.form is not None and self.item_held is None and self.resize_action is None and self.left_mouse_down() and self.value_change is None:
                self.resize_action = self.detect_rect_edge(self.form.x, self.form.y, self.form.width, self.form.height)
                if self.resize_action is not None and 0 <= self.resize_action <= 7:
                    for obj in self.group_selected:
                        obj.update_dictionary()
                        obj.is_selected = False
                    self.item_resized = self.form
                    on_other_item = False
                    for obj in self.form.objects:
                        if self.mouse_in_item(obj, is_on_form=True):
                            on_other_item = True
                    if not on_other_item:
                        self.group_selected = [self.form]
                    self.item_held = None
                elif self.resize_action is None and self.mouse_in_item(self.form):
                    for obj in self.group_selected:
                        obj.update_dictionary()
                        obj.is_selected = False
                    on_other_item = False
                    for obj in self.form.objects:
                        if self.mouse_in_item(obj, is_on_form=True):
                            on_other_item = True
                    if not on_other_item:
                        self.group_selected = [self.form]
                    self.item_held = self.form
                    self.item_resized = None
                for object in reversed(self.form.objects):
                    ra = self.detect_rect_edge(object.abs_x, object.abs_y, object.width, object.height)
                    if ra is not None:
                        self.resize_action = ra
                        self.item_resized = object
                        if object not in self.group_selected:
                            for obj in self.group_selected:
                                obj.update_dictionary()
                                obj.is_selected = False
                            self.group_selected = [object]
                        self.item_held = None
                        break
                    elif self.mouse_in_item(object, is_on_form=True):
                        if object not in self.group_selected:
                            for obj in self.group_selected:
                                obj.update_dictionary()
                                obj.is_selected = False
                            self.group_selected = [object]
                        self.item_held = object
                        self.item_resized = None
                        break
            elif self.form is not None and self.item_resized is not None:
                self.resize_item(self.item_resized)

            self.find_selected_group()
            for obj in self.group_selected:
                obj.is_selected = True

            if len(self.group_selected) != 0 and self.left_mouse_down() and self.value_change is None:
                for obj in self.group_selected:
                    diffx, diffy = pg.mouse.get_pos()[0] - self.prev_pos_left_down[0], pg.mouse.get_pos()[1] - self.prev_pos_left_down[1]
                    obj.x += diffx
                    obj.y += diffy
                    if diffx != 0 or diffy != 0:
                        self.has_changed = True
                self.prev_pos_left_down = pg.mouse.get_pos()

            if self.value_change is None:
                for obj in self.group_selected:
                    obj.update_dictionary()

            if self.item_held is not None:
                self.item_held.update_dictionary()

            self.update_screen()

    def load(self):
        loader = FormLoader(self.save_filepath, self.screen)
        loader.load()
        self.form = loader.form
        self.memory.remember(self.form.copy())
        self.exception_handler = loader.exception_handler

    def save(self):
        data_separator = '##END DATA##\n'
        to_save = []
        for (key, value) in [(key, value) for (key, value) in self.form.__dict__.items() if key in self.form.defaults]:
            to_save.append(f'{key}:{value}\n')
        to_save.append(data_separator)
        for object in self.form.objects:
            for (key, value) in [(key, value) for (key, value) in object.__dict__.items() if key in object.defaults]:
                if key == 'text':
                    value = value.replace('\n', '#')
                to_save.append(f'{key}:{value}\n')
            to_save.append(data_separator)
        savefile = open(self.save_filepath, 'w')
        savefile.writelines(to_save)
        savefile.close()